
//18F4620 SD CARD BOOTLOADER BETA //
/* Thanks Tolga Abaci */
/* convert 18f4620 10 mhz *4 hspll  Emre Zanbak */
/* All optimize enable and build = 1920 byte program memory - 839 kb data memory */

#include <p18f4620.h> 
#include <delays.h> 

//#include "lcd.h" 
//#include "utils.h"

//config du PIC

#pragma config OSC = HSPLL  // 10 Mhz x 4 40mhz
#pragma config FCMEN = OFF
#pragma config IESO = OFF
#pragma config PWRT = ON
#pragma config BOREN = OFF
#pragma config MCLRE = OFF
#pragma config WDT    = OFF  // watchdog off
#pragma config LPT1OSC = OFF
#pragma config PBADEN = OFF
#pragma config STVREN = ON
#pragma config LVP = OFF 
#pragma config XINST = OFF
#pragma config DEBUG = OFF
#pragma config CP0 = OFF
#pragma config CP1 = OFF
#pragma config CP2 = OFF
#pragma config CP3 = OFF
#pragma config CPB = OFF
#pragma config CPD = OFF
#pragma config WRT0 = OFF
#pragma config WRT1 = OFF
#pragma config WRT2 = OFF
#pragma config WRT3 = OFF
#pragma config WRTB = OFF
#pragma config WRTC = OFF
#pragma config WRTD = OFF


typedef signed char int8_t;
typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned long uint32_t;
typedef long int32_t;

#pragma udata
uint32_t fpp_begin_lba;
uint32_t fat_begin_lba;
uint32_t sectors_per_fat;
uint32_t cluster_begin_lba;
uint8_t sectors_per_cluster;
uint8_t fat_count;
uint16_t root_dir_first_cluster;
uint16_t reserved_sector_count;

#pragma udata sector_buffer_section
uint8_t sector_buffer[512];

#define RM_RESET_VECTOR          0x002000
#define RM_HIGH_INTERRUPT_VECTOR    0x002008 
#define RM_LOW_INTERRUPT_VECTOR     0x002018 

/* Vector remapping */
#pragma code _HIGH_INTERRUPT_VECTOR = 0x000008
void _high_ISR (void)
{
    _asm goto RM_HIGH_INTERRUPT_VECTOR _endasm
}

#pragma code _LOW_INTERRUPT_VECTOR = 0x000018
void _low_ISR (void)
{
    _asm goto RM_LOW_INTERRUPT_VECTOR _endasm
}

#pragma code
far rom char str_bin[] = "FIRMWARE BIN";

void erase_program_page(uint16_t address, uint8_t* buf)
{
   uint8_t i;
   far rom uint8_t* rom_ptr;
    rom_ptr = (far rom uint8_t*) address;
 
   /* Dummy read, required for loading the table pointer */
   i = *rom_ptr;

   /* Erase */
    EECON1bits.EEPGD = 1;     
    EECON1bits.CFGS = 0;      
    EECON1bits.FREE = 1;      
    EECON1bits.WREN = 1;      
    INTCONbits.GIE = 0;     
    EECON2 = 0x55;          
    EECON2 = 0xAA; 
    EECON1bits.WR = 1;      
    INTCONbits.GIE = 1;                   

   /* Write 64 bytes */
    for (i = 0; i < 64; i++) 
        *rom_ptr++ = buf[i];    

    EECON1bits.EEPGD = 1;               
    EECON1bits.CFGS = 0;                
    EECON1bits.FREE = 0;                
    EECON1bits.WREN = 1;                
    INTCONbits.GIE = 0;     
    EECON2 = 0x55;          
    EECON2 = 0xAA; 
    EECON1bits.WR = 1;      
    INTCONbits.GIE = 1;     
    EECON1bits.WREN = 0;                          
}



#define CARD_TYPE_SD 0
#define CARD_TYPE_SDHC 1

uint8_t card_type;

/* SPI Send/Receive byte */
static uint8_t SPI(uint8_t d) 
{  
   char received = 0;
   
   SSPBUF = d;
   while (SSPSTATbits.BF != 1);      
   received = SSPBUF;
   
   return received;
}

/* Send a command to SD/MMC */
uint8_t card_command(uint8_t cmd, uint16_t addrH, uint16_t addrL, uint8_t crc)
{   
   SPI(0xFF);
   SPI(cmd);
   SPI((uint8_t) (addrH >> 8));
   SPI((uint8_t) addrH);
   SPI((uint8_t) (addrL >> 8));
   SPI((uint8_t) addrL);
   SPI(crc);
   SPI(0xFF);   
   return SPI(0xFF);   // return the last received character
}

/* Initalize SD/MMC Card */
uint8_t card_init(void) 
{ 
   uint8_t i;

   /* Slow clock during card identification */
   TRISCbits.TRISC3 = 0;   
   TRISCbits.TRISC5 = 0;   
   SSPCON1bits.SSPEN = 0;
   SSPCON1 = 0x12;         // CKP High, SPI Master, clock = Fosc/64
   SSPSTATbits.SMP = 0;
   //SSPSTATbits.SMP = 1;
   SSPSTATbits.CKE = 0;
   SSPCON1bits.SSPEN = 1;
   
   /* Raise Chip Select */
   TRISCbits.TRISC0 = 0;
   PORTCbits.RC0 = 1;   
   
   /* Switch card to SPI Mode */
   for(i = 0; i < 10; i ++) 
      SPI(0xFF); 
      
   /* Lower Chip Select */
   PORTCbits.RC0 = 0;

   /* Send CMD0 */
   if (card_command(0x40, 0x0000, 0x0000, 0x95) != 0x01) 
      return 0;

   card_type = CARD_TYPE_SD;   
   if ((card_command(0x48, 0x0000, 0x1AA, 0x87) & 4) == 0)   
   {
      SPI(0xFF);
      SPI(0xFF);      
      if ((SPI(0xFF) & 1) == 0)
         return 0;
      if (SPI(0xFF) != 0xAA)
         return 0;
         
      /* Send ACMD41 */    
      while (1)
      {
         uint8_t result;

         card_command(0x77, 0x0000, 0x0000, 0xFF);
         result = card_command(0x69, 0x4000, 0x0000, 0xFF);
         if (result == 0x00)
            break;                        
      }
   
      /* Send CMD58 */
      if (card_command(0x7A, 0x0000, 0x0000, 0xFF))
         return 0;         
        
      if(SPI(0xFF) & 0x40)
         card_type = CARD_TYPE_SDHC;

      SPI(0xFF);
      SPI(0xFF);
      SPI(0xFF);                     
   }                     
   else
   {   
      /* Send CMD1 */    
      while (card_command(0x41, 0x0000, 0x0000, 0xFF) != 0x00);
   }

   /* Raise Chip Select */
   PORTCbits.RC0 = 1;
   
   /* Speed up the clock */
   SSPCON1bits.SSPEN = 0;
   //SSPCON1 = 0x11;         // CKP High, SPI Master, clock = Fosc/16
   SSPCON1 = 0x10;         // CKP High, SPI Master, clock = Fosc/4
   SSPSTATbits.SMP = 0;
   //SSPSTATbits.SMP = 1;
   SSPSTATbits.CKE = 0;
   SSPCON1bits.SSPEN = 1;
   
   return 1;
}

/* Reads one sector of 512 bytes from the card */
uint8_t card_read(uint32_t sector, uint8_t* buffer) 
{
   uint16_t i;
   uint8_t result;
   uint8_t card_response; 
   
   uint32_t address = (card_type == CARD_TYPE_SDHC) ? (sector) : (sector << 9);      
      
   /* Lower Chip Select */
   PORTCbits.RC0 = 0;

   /* Send CMD17 */
   result = card_command(0x51, (uint16_t) (address >> 16), (uint16_t) address, 0xFF);   
            
   while (result != 0)
      result = SPI(0xFF);
         
   /* Wait until 0xFE is received */
   while ((card_response = SPI(0xFF)) == (uint8_t) 0xFF);
      
   if (card_response != 0xFE)   
      return 0;                  
   
   /* Read data */
   for(i = 0; i < 512; i ++) 
      buffer[i] = SPI(0xFF);  
         
   /* Receive CRC */
   SPI(0xFF); 
   SPI(0xFF);
      
   /* Let the card finish */
   SPI(0xFF); 
         
   return 1;   
}

#define LE16(card_sector, X)   (((uint16_t) card_sector[(X)]) +          \
                         (((uint16_t) card_sector[(X) + 1]) << 8))    
             
#define LE32(card_sector, X)   (((uint32_t) card_sector[(X)]) +          \
                        (((uint32_t) card_sector[(X) + 1]) << 8) +     \
                        (((uint32_t) card_sector[(X) + 2]) << 16) +    \
                        (((uint32_t) card_sector[(X) + 3]) << 24))      
             
#define LBA(X) (cluster_begin_lba + ((X) - 2) * sectors_per_cluster)

/* Initializes FAT32 support */
uint8_t fat32_init(void)
{
   uint16_t bytes_per_sector;   

   /* Read sector 0 into the sector buffer */
   if (!card_read(0, sector_buffer))
      return 0;

   /* Depending on whether SD Card controller firmware reports 
    * itself as removable media or not, it may or may not have an
    * MBR. If there is no MBR, sector 0 will be the boot sector, as
    * in a floppy. If the first byte of sector 0 is 0xEB, we assume
    * it's the boot sector (LBR).
    */
          
   if (sector_buffer[0] != 0xEB)
   {
      /* Find out the starting sector of the first primary partition */ 
      fpp_begin_lba = LE32(sector_buffer, 454);      

      /* Read FAT32 Volume ID into the sector buffer */
      if (!card_read(fpp_begin_lba, sector_buffer))
         return 0;
   }
   else
      fpp_begin_lba = 0;             
   
   /* Compute important FAT locations */
   reserved_sector_count = LE16(sector_buffer, 14);
   fat_count = sector_buffer[16];
   sectors_per_fat = LE32(sector_buffer, 36);
   
   fat_begin_lba = fpp_begin_lba + reserved_sector_count;
   cluster_begin_lba = fpp_begin_lba + reserved_sector_count + (fat_count * sectors_per_fat);
   
   sectors_per_cluster = sector_buffer[13];   
   root_dir_first_cluster = LE32(sector_buffer, 44);

   return 1;
}

/* Looks for FIRMWARE.BIN in the SD Card */
void update_firmware(void)
{
   uint8_t i;
   uint8_t* dir_entry;
   uint32_t cluster;
   uint32_t size;
   uint16_t address;
   int32_t bytes_left;
   uint32_t sector;

   if (!card_init())
      return;         
   if (!fat32_init())
      return;   

   /* Read directory first sector into the sector buffer */
   if (!card_read(LBA(root_dir_first_cluster), sector_buffer))
      return;   

   size = 0;
   dir_entry = sector_buffer;
   
   while (dir_entry[0])
   {
      for (i = 0; i < 11; i ++)
         if (dir_entry[i] != str_bin[i])
            break;
      
      if (i == 11)
      {
         /* Get the cluster number */
         cluster = ((uint32_t) dir_entry[26]) + 
                 ((uint32_t) dir_entry[27] << 8) + 
                 ((uint32_t) dir_entry[20] << 16) + 
                 ((uint32_t) dir_entry[21] << 24);   
                     
         /* Get the size */
         size = LE32(dir_entry, 28);

         break;
      }

      dir_entry += 0x20;
   }
                                    
   if (size)
   {
   
                  //   LCDsetcursorposition(0x00); 
               //      lcd_print((const far rom char*) "Updating firmware");
  
    
      address = 0;
      bytes_left = size;
      sector = LBA(cluster);
      
      while (bytes_left > 0)
      {      
         if (!card_read(sector, sector_buffer))
            return;
      
         for (i = 0; i < 8; i ++)
         {
            if (bytes_left <= 0)      
               break;
      
            if (address >= 0x2000)
            {
            
               //   LCDsetcursorposition(0x00); 
               //      lcd_print((const far rom char*) "Writing Block", address);
               
               erase_program_page(address, sector_buffer + ((uint16_t) i) * 64);
            }

            address += 64;
            bytes_left -= 64;
      
            //Delay10KTCYx(20);
         }

   //      blink led message
      
       
         sector ++;
      }
   }   
   else
Nop();
//   LCDsetcursorposition(0x40); 
//      lcd_print((const far rom char *)"Firmw. not found");        
}




   
void main(void) 
{

   /* No Analog Pins */      
   ADCON1 |= 0x0F;                 // Default all pins to digital


   INTCON2bits.RBPU=0;     // port B pull up enabled


//   test();

//   PORTD=0x08;
//   TRISD=0xF0; // lcd data bus input , led out, lcd ctrl out
//initLCD();
   /* Update firmware? */
//LCDsetcursorposition(0x00);   
//lcd_print((const far rom char *)"-------------- ");
Delay10KTCYx(200);
   if (PORTEbits.RE2 == 0)
   {

      update_firmware();     
	 while (1); 
   }

   _asm goto RM_RESET_VECTOR _endasm
} 


